/*
   _===_ _           _                  ______ _     _             _ _ _
  |  ___| |         | |                 | ___ (_)   | |           (_) | |
  | |__ | | ___  ___| |_ _ __ ___ ______| |_/ /_  __| | ___  _   _ _| | | ___ _   _ _ __
  |  __|| |/ _ \/ __| __| '__/ _ \______| ___ \ |/ _` |/ _ \| | | | | | |/ _ \ | | | '__|
  | |___| |  __/ (__| |_| | | (_) |     | |_/ / | (_| | (_) | |_| | | | |  __/ |_| | |
  \____/|_|\___|\___|\__|_|  \___/      \____/|_|\__,_|\___/ \__,_|_|_|_|\___|\__,_|_|

 *  
 * Fréquencemètre 150 MHz.
 * File:    Compteur_8_Chiffres_Verilog.v
 * Author:  Bertrand, Électro-Bidouilleur, electro@bidouilleur.ca
 * Created: 15 Février 2019
 * Target:  Plaquette de développement A-C4E6 (FPGA Intel Cyclone IV - EP4CE6E22C8)
 * Environment: Quartus Prime Lite
*/
 
module Compteur_8_Chiffres_Verilog  (clk_50M,digit_sel,freq_in,dataout,freq_out,ref_in,ref_select,overflow_led);

// Les broches d'entrées et soties
input clk_50M;     					// Entrée du Signal de l'Oscillateur 50 MHZ
input freq_in;						// Entrée du signal à mesurer
input ref_in;						// Entrée du signal de référence externe 10 MHz
input ref_select;					// Entrée Interrupteur de sélection de référence (Osc 50 MHz ou référence externe 10 MHz)
output[6:0] dataout;   				// Sorties des Segments "a" à "g" des affichages à LED
output[7:0] digit_sel;				// Sorties qui allument les chiffres du fréquencemètre, un à la fois
output freq_out;					// sortie du Signal de mesure de l'horloge interne (assigné sur le 1 Hz vers la LED)
output overflow_led;				// Sortie de la LED de débordement de l'affichage (pour > 100 MHz)

// Les registres
reg freq_out;						// Signal de mesure de l'horloge interne (assigné sur le 1 Hz)
wire freq_in;						// Signal à mesurer
wire ref_in;						// Signal de référence externe 10 MHz
wire ref_select;					// Interrupteur de sélection de référence (Osc 50 MHz ou référence externe 10 MHz)
reg clk_50M_1Hz;					// Horloge 1Hz dérivée de l'oscillateur 50 MHz
reg clk_10M_1Hz;					// Horloge 1Hz dérivée de la référence externe 10 MHz
reg clk_1Hz;						// Horloge 1Hz sélectionnée pour la base de temps (gate)
reg[25:0] counter_50M = 0;     		// Compteur incrémentant à 50 MHZ - Valeur maximale de 50 millions ( < 2^26)
reg[23:0] counter_10M = 0;     		// Compteur incrémentant à 10 MHZ - Valeur maximale de 10 millions ( < 2^24)
reg[27:0] counter_FreqIn = 0;		// Compteur du signal d'entrée - Valeur maximale de 268 millions ( < 2^28)
reg[27:0] latched_cntr;      		// Registre de capture courante du compteur de signal d'entrée - Valeur maximale de 268 millions ( < 2^28)
reg[27:0] old_latched_cntr = 0; 	// Registre de capture précédente du compteur de signal d'entrée - Valeur maximale de 268 millions ( < 2^28)
reg[27:0] latched_reading;			// Fréquence lue (en Hz) tirée des registres de capture de fréquence - Valeur maximale de 268 millions ( < 2^28)
reg[31:0] latched_reading_BCD;		// Valeur BCD (forme binaire) de la fréquence à afficher (les 8 chiffres sont concatenés) - 4 bits x 8 chiffres = 32 bits
reg clk_mux = 0;					// Horloge de multiplexage de l'affichage
reg[6:0] dataout;					// Segments "a" à "g" des affichages à LED (connectés en commun sur tous les chiffres)
reg[3:0] BCD = 0;					// Valeur BCD du chiffre courant à afficher
reg[7:0] digit_sel = 8'b11111110;  	// Lignes qui sélectionnent, un à la fois, les chiffres à afficher
reg[5:0] i;							// Variable compteur de boucle
reg overflow_led;					// LED de débordement de l'affichage (pour > 100 MHz)

// Les constantes
parameter DIVISOR50 = 50000000; 	// Constante pour diviser un compteur par 50 millions
parameter DIVISOR10 = 10000000; 	// Constante pour diviser un compteur par 10 millions

// La logique débute ici

// Logique active lors du changement de ref_select
always  @ ( ref_select )
 begin
  if (ref_select == 1) clk_1Hz = clk_50M_1Hz;			// Sélection de la référence 1 Hz utilisée d'après l'interrupteur de sélection
  else clk_1Hz = clk_10M_1Hz;							//    "
 end
 
// Logique activée sur les fronts montants de l'horloge 50 MHz
always @ ( posedge clk_50M )
 begin
  freq_out  = !clk_1Hz;									// Le signal de DEL est la base de temps de 1 Hz (inversée)
  // Gestion du compteur incrémentant à 50 MHz
  counter_50M = counter_50M + 1;						// Incrémentation du compteur
  if(counter_50M > (DIVISOR50-1)) counter_50M = 0;		// Maximum 49999999, sinon remise à zéro

  // Production de la référence 1 Hz dérivée de l'oscillateur 50 MHz
  if (counter_50M < (DIVISOR50/2)) clk_50M_1Hz = 0;		// À moins de 25000000 on met 0
  else clk_50M_1Hz = 1;									// À 25000000 et plus on met 1

  // Production de l'horloge 1524 Hz de multiplexage de l'affichage
  clk_mux = counter_50M[14];  							// Horloge de multiplexage créée avec le bit 14 du compteur 50 MHz
 end

// Logique activée sur les fronts montants de la référence de 10 MHz
always @ ( posedge ref_in )
 begin
  // Gestion du compteur incrémentant à 10 MHz
  counter_10M = counter_10M + 1;						// Incrémentation du compteur
  if(counter_10M > (DIVISOR10-1)) counter_10M = 0;		// Maximum 9999999, sinon remise à zéro
  
  // Production de l'horloge 1 Hz dérivée de l'oscillateur 10 MHz
  if (counter_10M < (DIVISOR10/2)) clk_10M_1Hz = 0;		// À moins de 5000000 on met 0
  else clk_10M_1Hz = 1;									// À 5000000 et plus on met 1
 end

// Logique activée sur les fronts montants de l'entrée de fréquence
always @ ( posedge freq_in )
 begin
  // Gestion du compteur incrémentant par l'entrée à mesurer
  counter_FreqIn = counter_FreqIn + 1;					// Incrémenter le compteur d'entrée (on le laisse déborder au besoin)
 end
 
// Logique activée sur les fronts montants de l'horloge de multiplexage de l'affichage
 always @ ( posedge clk_mux ) 
 begin
   // gestion du multiplexage des 8 chiffres de l'affichage
  digit_sel = (digit_sel << 1) + 1;							// Sélectionner le chiffre d'affichage suivant (shift left, logique inversée)
  if (digit_sel == 8'b11111111) digit_sel = 8'b11111110;	// Si tous les chiffres on tété affichés, on recommence avec le premier chiffre

  // Transfert de la bonne valeur en BCD selon le chiffre à afficher
  case (digit_sel)
   8'b11111110: BCD = latched_reading_BCD[3:0];   	// Valeur en BCD tirée des groupes de 4 bits selon le chiffre à afficher
   8'b11111101: BCD = latched_reading_BCD[7:4];		//      "				
   8'b11111011: BCD = latched_reading_BCD[11:8];	//      "				
   8'b11110111: BCD = latched_reading_BCD[15:12];	//      "				
   8'b11101111: BCD = latched_reading_BCD[19:16];	//      "				
   8'b11011111: BCD = latched_reading_BCD[23:20];	//      "				
   8'b10111111: BCD = latched_reading_BCD[27:24];	//      "				
   8'b01111111: BCD = latched_reading_BCD[31:28];	//      "				
  endcase

  // Décodeur de BCD à affichage 7 segments
  case (BCD)					
  0: dataout = 7'b1000000;		// Conversion BCD à combinaison des 7 segments LED formant un chiffre  
  1: dataout = 7'b1111001;		//      "				
  2: dataout = 7'b0100100;		//      "				  
  3: dataout = 7'b0110000;		//      "				
  4: dataout = 7'b0011001;		//      "				
  5: dataout = 7'b0010010;		//      "				  
  6: dataout = 7'b0000010;		//      "				 
  7: dataout = 7'b1111000;		//      "				
  8: dataout = 7'b0000000;		//      "				
  9: dataout = 7'b0010000;		//      "				
  endcase
 end

// Logique activée par un front montant de la base de temps de 1 Hz (calcul de la fréquence à afficher)
 always @ ( posedge clk_1Hz )
 begin
  latched_cntr = counter_FreqIn;										// Capturer la valeur courante du compteur de signal d'entrée
  latched_reading = latched_cntr - old_latched_cntr;					// Soustraire l'ancienne valeur de la nouvelle pour dériver la fréquence en Hz
  old_latched_cntr = latched_cntr;										// La valeur courante devient l'ancienne valeur
  if (latched_reading < 0) latched_reading = latched_reading + 1<<28;	// Tenir compte d'un débordement de compteur dans le calcul de la fréquence en Hz
  if (latched_reading  > 99999999) overflow_led = 0;					// Allumer la LED de déborfdement d'affichage si la fréquence >= 100000000 Hz
  else overflow_led = 1;												// Sinon éteindre la LED de débordement

   // Algorithme "Double Dabble" de conversion binaire à BCD (par groupe de 4 bits concatenés en BCD)
  latched_reading_BCD = 0; 				// Remise à zéro de la variable 32 bits d'affichage en BCD.
  for (i = 0; i < 28; i = i+1) 			// Boucler pour couvrir les 28 bits de la fréquence en binaire
  begin
   latched_reading_BCD = {latched_reading_BCD[30:0],latched_reading[27-i]}; // concaténation constructive des 8 chiffres en BCD
   if(i < 27 && latched_reading_BCD[3:0] > 4) 								// Si un chiffre en binaire vaut plus de 4, y ajouter 3
   latched_reading_BCD[3:0] = latched_reading_BCD[3:0] + 3;					//      "
   if(i < 27 && latched_reading_BCD[7:4] > 4)								//      "
   latched_reading_BCD[7:4] = latched_reading_BCD[7:4] + 3;					//      "
   if(i < 27 && latched_reading_BCD[11:8] > 4)								//      "
   latched_reading_BCD[11:8] = latched_reading_BCD[11:8] + 3;  				//      "
   if(i < 27 && latched_reading_BCD[15:12] > 4)								//      "
   latched_reading_BCD[15:12] = latched_reading_BCD[15:12] + 3;				//      "  
   if(i < 27 && latched_reading_BCD[19:16] > 4)								//      "
   latched_reading_BCD[19:16] = latched_reading_BCD[19:16] + 3;				//      "  
   if(i < 27 && latched_reading_BCD[23:20] > 4)								//      "
   latched_reading_BCD[23:20] = latched_reading_BCD[23:20] + 3;				//      "  
   if(i < 27 && latched_reading_BCD[27:24] > 4)								//      "
   latched_reading_BCD[27:24] = latched_reading_BCD[27:24] + 3;				//      "  
   if(i < 27 && latched_reading_BCD[31:28] > 4)								//      "
   latched_reading_BCD[31:28] = latched_reading_BCD[31:28] + 3;				//      "  
  end   
 end
endmodule
